/*
 * File:	DemoI2C.cpp
 *
 * Date:	2012-08-21
 * Company:	DH electronics GmbH
 * Author:	Ludwig Zenz, lzenz@dh-electronics.de
 *
 * Remark:	The following code aims to offer you some help
 * 			with the i2c operations in Linux user-space programs.
 */


/*
 *	Key to symbols which are used at the protocol descriptions
 *	----------------------------------------------------------
 *
 *	S     (1 bit) : Start bit
 *	P     (1 bit) : Stop bit
 *	Rd/Wr (1 bit) : Read/Write bit. Rd equals 1, Wr equals 0.
 *	A, NA (1 bit) : Accept and reverse accept bit.
 *	Addr  (7 bits): I2C 7 bit address. Note that this can be expanded as usual to
 *	                get a 10 bit I2C address.
 *	Comm  (8 bits): Command byte, a data byte which often selects a register on
 *	                the device. (In our case the eeprom-addr)
 *	Data  (8 bits): A plain data byte.
 *
 *	[..]: Data sent by I2C device, as opposed to data sent by the host adapter.
 */


/*
 * For Further i2c documentation and examples have a look at:
 * ->	http://www.kernel.org/doc/Documentation/i2c/dev-interface
 * ->	http://www.elinux.org/Interfacing_with_I2C_Devices
 */


/* Header */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>

/* I2C Header */
#include <linux/i2c.h>
#include <linux/i2c-dev.h>


/* defines */
#define BUFFERSIZE 		20
#define I2C_ADAPTER		0x03
#define I2C_DEV_ADDR	0x50

/* prototypes */
int open_a_devicefile(int adapter_nr);

int byte_write(int file, unsigned char *buf, int length);
int byte_read(int file, unsigned char *buf, int length);

int block_write(int file,int dev_addr,int eeprom_addr,unsigned char *buf, int length);
int block_read(int file,int dev_addr,int eeprom_addr,unsigned char *buf, int length);


int main()
{
	int file;
	int dev_addr = I2C_DEV_ADDR; 	/* The I2C address */

	unsigned char buf[BUFFERSIZE];

	memset(buf, 0, BUFFERSIZE);

	printf("\nStarting I2C demo ...\n\n");

	/*************************************************************************************
	 * The first thing to do is to open a device-file
	 */
	file = open_a_devicefile(I2C_ADAPTER);


	/*************************************************************************************
	 * Specify the device address to communicate with
	 * -> (this is necessary if you want to use read() and write() )
	 */
	if (ioctl(file, I2C_SLAVE, dev_addr) < 0)
	{
		printf("ERROR during setup the device i2c-address!!\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	/*************************************************************************************
	 * read some bytes from EEPROM address 0x10 with write() and read() in combination
	 */
	buf[0] = 0x10;

	/* setup eeprom-address from which we want to read */
	byte_write(file, buf, 1);

	/* read 5 bytes from the address configured before */
	byte_read(file, buf, 5);

	printf("\nbyte-read:\n");
	for (int index = 0; index < 5; index++)
	{
		printf("buf[%d]=%02X  \n",index, buf[index]);
	}

	/*************************************************************************************
	 * do a block write to eeprom-addr 0x20 (20 bytes)
	 */

	for (int index = 0; index < 20; index++)
	{
		buf[index] = 0xFF - index;
	}

	block_write(file, dev_addr, 0x20, buf, 20);

	/* clear buffer */
	memset(buf, 0, BUFFERSIZE);

	/* wait some time until the eeprom is ready again */
	sleep(1);

	/*************************************************************************************
	 * do a block read from eeprom-addr 0x20 (20 bytes)
	 */

	block_read(file, dev_addr, 0x20, buf, 20);

	printf("\nblock-read:\n");
	for (int index = 0; index < 20; index++)
	{
		printf("buf[%d]=%02X  \n",index, buf[index]);
	}

	/*************************************************************************************
	 * close the device-file
	 */
	close(file);

	return 0;
}


/************************************************************************************************
 * low level functions
 ************************************************************************************************/


/*
 * Open a device-file to an i2c-adapter
 */
int open_a_devicefile(int adapter_nr)
{
	int file;
	char filename[20];

	snprintf(filename, 19, "/dev/i2c-%d", adapter_nr);
	file = open(filename, O_RDWR);

	/* Error Handling */
	if (file < 0)
	{
		printf("ERROR opening %s!!\n", filename);
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}
	else
		printf("\nOpen %s was successful!!\n\n", filename);

	return file;
}


/*
 * write some bytes
 *
 * used protocol:
 * S Addr Wr [A] Data [A] Data [A] ... Data [A] P
 *
 * Remark: You have to setup the i2c-addr with "ioctl(file, I2C_SLAVE, dev_addr)" first once!
 */
int byte_write(int file, unsigned char *buf, int length)
{
	if ( write(file, buf, length) != length)
	{
		printf("ERROR: byte_write() failed!!!\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	return 0;
}


/*
 * read some bytes
 *
 * used protocol:
 * S Addr Rd [A] [Data] A [Data] A ... A [Data] NA P
 *
 * Remark: You have to setup the i2c-addr with "ioctl(file, I2C_SLAVE, dev_addr)" first once!
 */
int byte_read(int file, unsigned char *buf, int length)
{

	if (read(file, buf, length) != length)
	{
		printf("ERROR: byte_read() failed!!!\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	return 0;
}

/*
 * block write
 *
 * used protocol:
 * S Addr Wr [A] Comm [A] Data [A] Data [A] ... [A] Data [A] P
 */
int block_write(int file,int dev_addr,int eeprom_addr,unsigned char *buf,int length)
{
	int err = 0;

	struct i2c_msg msg[1];
	struct i2c_rdwr_ioctl_data msgst;

	unsigned char *buffer;
	buffer = (unsigned char *)malloc(1 + length);

	buffer[0] = eeprom_addr;
	memcpy(buffer + 1, buf ,length);

	/* setup eeprom-address msg */
	msg[0].addr = dev_addr;
	msg[0].flags = 0;
	msg[0].len = 1 + length;
	msg[0].buf = (__u8*)buffer;

	msgst.msgs = msg;
	msgst.nmsgs = 1;

	err = ioctl(file,I2C_RDWR,&msgst); 		/* Combined R/W transfer (one STOP only) */
	if (err < 0)
	{
		printf("ERROR: block_write() failed!!!\n");
		free(buffer);
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	free(buffer);
	return 0;
}


/*
 * block read (one STOP only)
 *
 * used protocol:
 * S Addr Wr [A] Comm [A]  S Addr Rd [A] [Data] A [Data] A ... A [Data] NA P
 */
int block_read(int file,int dev_addr,int eeprom_addr,unsigned char *buf, int length)
{
	int err = 0;
	char eeprom_addr_buf[2];

	struct i2c_msg msg[2];
	struct i2c_rdwr_ioctl_data  msgst;

	eeprom_addr_buf[0] = eeprom_addr;

	/* setup eeprom-address msg */
	msg[0].addr = dev_addr;
	msg[0].flags = 0;
	msg[0].len = 1;
	msg[0].buf = (__u8*)eeprom_addr_buf;

	/* setup data-to-read msg */
	msg[1].addr = dev_addr;
	msg[1].flags = I2C_M_RD; 				/* Read Data from Slave to Master */
	msg[1].len = length;
	msg[1].buf = buf;

	msgst.msgs = msg;
	msgst.nmsgs = 2;

	err = ioctl(file, I2C_RDWR, &msgst);	/* Combined R/W transfer (one STOP only) */

	if (err < 0)
	{
		printf("ERROR: block_read() failed!!!\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	return err;
}


