/*
 * File:	DemoSPI.cpp
 *
 * Date:	2012-08-22
 * Company:	DH electronics GmbH
 * Author:	Ludwig Zenz, lzenz@dh-electronics.de
 *
 * Remark:	The following code aims to offer you some help with the
 * 			SPI operations in Linux	user-space programs.
 *
 * 			In this example we write and read to a spi-eeprom (Microchip 25AA010A)
 */


/*
 * For further SPI documentation and examples have a look at:
 * ->	http://www.kernel.org/doc/Documentation/spi/spidev
 * ->	additional demos/spidev_test.c
 * ->	additional_demos/spidev_fdx.c
 */

/* IO Header */
#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/types.h>

/* SPI Header */
#include <linux/spi/spidev.h>
#include "spi_eeprom_25AA010A.h"

/* defines */
#define BITS_SPIWORD	8
#define SPEED_IN_HZ		3000000 /* 3 MHz */

/* available interfaces on am35x */
#define DHCOM_SPI1		1
#define DHCOM_SPI1_CS	1

#define DHCOM_SPI2		2
#define DHCOM_SPI2_CS	0



/* prototypes */
int open_a_devicefile(int intf_nr, int cs_nr);
int configure_spi_interface(int file, uint8_t mode, uint8_t bits, uint32_t speed_Hz);
int do_spi_transfer(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay);

int read_sequence(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay, uint8_t eeprom_addr, uint8_t *rx_buffer, int length);
int write_sequence(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay, uint8_t eeprom_addr, uint8_t *tx_buffer, int length);

int enable_writing(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay);
int disable_writing(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay);

int write_status_reg(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay, uint8_t status_reg);
int read_status_reg(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay, uint8_t *status_reg);

int main()
{
	int ret = 0;
	int file;
	uint8_t status_reg = 0;
	int length = 16;		/* max 16 bytes in a sequence (depending on the used eeprom)*/

	printf("\nStarting SPI demo ...\n\n");

	/*************************************************************************************
	 * The first thing to do is to open a device-file
	 */
	file = open_a_devicefile(DHCOM_SPI1, DHCOM_SPI1_CS);

	/*************************************************************************************
	 * configure the spi-interface
	 */
	configure_spi_interface(file, SPI_MODE_0, BITS_SPIWORD, SPEED_IN_HZ);

	/*************************************************************************************
	 * writing some data
	 */

	unsigned char *tx_buffer = (unsigned char *)malloc(length);

	for (int index = 0; index < length; index++)
	{
		tx_buffer[index] = index + 1;
	}

	enable_writing(file, BITS_SPIWORD, SPEED_IN_HZ, 0);

	read_status_reg(file, BITS_SPIWORD, SPEED_IN_HZ, 0, &status_reg);
	if (!(status_reg & (STATUS_WRITE_EN_LATCH | STATUS_WRITE_IN_PROGRESS)))
	{
		/* wait until WP is disabled */
		usleep(500000);
	}

	write_sequence(file, BITS_SPIWORD, SPEED_IN_HZ, 0, 0x10, tx_buffer, length);

	read_status_reg(file, BITS_SPIWORD, SPEED_IN_HZ, 0, &status_reg);

	if (status_reg & STATUS_WRITE_IN_PROGRESS)
	{
		/* wait some time for eeprom to finish writing */
		usleep(500000);
	}

	/*************************************************************************************
	 * reading some data
	 */
	unsigned char *rx_buffer = (unsigned char *)malloc(length);

	read_sequence(file, BITS_SPIWORD, SPEED_IN_HZ, 0, 0x10, rx_buffer, length);


	for (int index = 0; index < length; index++)
	{
		if (!(index % 6)) puts("");

		printf("%.2X ", rx_buffer[index]);
	}
	puts("");

	close(file);

	return ret;
}

/*
 * open spi-device-file
 */
int open_a_devicefile(int intf_nr, int cs_nr)
{
	int file;
	char filename[15];

	snprintf(filename, 15, "/dev/spidev%d.%d", intf_nr, cs_nr);

	file = open(filename, O_RDWR);

	/* Error Handling */
	if (file < 0)
	{
		printf("ERROR opening %s!!\n", filename);
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}
	else
		printf("\nOpen %s was successful!\n", filename);

	return file;
}

/*
 * configure the spi interface
 */
int configure_spi_interface(int file, uint8_t mode, uint8_t bits, uint32_t speed_Hz)
{
	int ret = 0;

	/* spi mode	 */
	ret = ioctl(file, SPI_IOC_WR_MODE, &mode);
	if (ret == -1)
	{
		printf("ERROR: can't set spi mode\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	ret = ioctl(file, SPI_IOC_RD_MODE, &mode);
	if (ret == -1)
	{
		printf("ERROR: can't get spi mode\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	/* bits per word */
	ret = ioctl(file, SPI_IOC_WR_BITS_PER_WORD, &bits);
	if (ret == -1)
	{
		printf("ERROR: can't set bits per word\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	ret = ioctl(file, SPI_IOC_RD_BITS_PER_WORD, &bits);
	if (ret == -1)
	{
		printf("ERROR: can't get bits per word\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	/* max speed Hz */
	ret = ioctl(file, SPI_IOC_WR_MAX_SPEED_HZ, &speed_Hz);
	if (ret == -1)
	{
		printf("ERROR: can't set max speed hz\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	ret = ioctl(file, SPI_IOC_RD_MAX_SPEED_HZ, &speed_Hz);
	if (ret == -1)
	{
		printf("ERROR: can't get max speed hz\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	printf("\n\nspi mode: %d\n", mode);
	printf("bits per word: %d\n", bits);
	printf("max speed: %d Hz (%d KHz)\n", speed_Hz, speed_Hz/1000);

	return 0;
}

/*
 * read a sequence of bytes
 */
int read_sequence(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay, uint8_t eeprom_addr, uint8_t *rx_buffer, int length)
{
	int ret;

	struct spi_ioc_transfer tr;
	uint8_t *tx;
	uint8_t *rx;

	/* add 2 bytes for length and eeprom-addr */
	length += 2;

	tx = (uint8_t  *)malloc(length);
	rx = (uint8_t  *)malloc(length);

	tx[0] = CMD_READ; 	 /* read command */
	tx[1] = eeprom_addr; /* set eeprom-addr */

	/* setup spi transfer struct */
	tr.tx_buf = (unsigned long)tx;
	tr.rx_buf = (unsigned long)rx;
	tr.len = length;
	tr.delay_usecs = delay;
	tr.speed_hz = speed_Hz;
	tr.bits_per_word = bits;


	ret = ioctl(file, SPI_IOC_MESSAGE(1), &tr);
	if (ret < 1)
	{
		printf("ERROR: can't send spi message\n");
		free(tx);
		free(rx);
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	for (int index = 2; index < length; index++)
	{
		rx_buffer[index-2] = rx[index];
	}

	free(tx);
	free(rx);

	return 0;
}

/*
 * write some bytes
 */
int write_sequence(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay, uint8_t eeprom_addr, uint8_t *tx_buffer, int length)
{
	int ret;

	struct spi_ioc_transfer tr;
	uint8_t *tx;
	uint8_t *rx;

	/* add 2 bytes for length and eeprom-addr */
	length += 2;

	tx = (uint8_t  *)malloc(length);
	rx = (uint8_t  *)malloc(length);

	tx[0] = CMD_WRITE; 		/* write command */
	tx[1] = eeprom_addr; 	/* set eeprom-addr */

	/* copy data into tx buffer */
	for (int index = 2; index < length; index++)
	{
		tx[index] =tx_buffer[index-2];
	}

	/* setup spi transfer struct */
	tr.tx_buf = (unsigned long)tx;
	tr.rx_buf = (unsigned long)rx;
	tr.len = length;
	tr.delay_usecs = delay;
	tr.speed_hz = speed_Hz;
	tr.bits_per_word = bits;


	ret = ioctl(file, SPI_IOC_MESSAGE(1), &tr);
	if (ret < 1)
	{
		printf("ERROR: can't send spi message\n");
		free(tx);
		free(rx);
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	free(tx);
	free(rx);

	return 0;
}

/*
 * Set the write enable latch (enable write operations)
 */
int enable_writing(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay)
{
	int ret;
	uint8_t tx;
	uint8_t rx;
	struct spi_ioc_transfer tr;


	tx = CMD_WRITE_EN;	/* command to enable writing */

	/* setup spi transfer struct */
	tr.tx_buf = (unsigned long)&tx;
	tr.rx_buf = (unsigned long)&rx;
	tr.len = 1;
	tr.delay_usecs = delay;
	tr.speed_hz = speed_Hz;
	tr.bits_per_word = bits;


	ret = ioctl(file, SPI_IOC_MESSAGE(1), &tr);
	if (ret < 1)
	{
		printf("ERROR: can't send spi message\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	return 0;
}

/*
 * Reset the write enable latch (disable write operations)
 */
int disable_writing(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay)
{
	int ret;
	uint8_t tx;
	uint8_t rx;
	struct spi_ioc_transfer tr;


	tx = CMD_WRITE_DIS;		/* command to disable writing */

	/* setup spi transfer struct */
	tr.tx_buf = (unsigned long)&tx;
	tr.rx_buf = (unsigned long)&rx;
	tr.len = 1;
	tr.delay_usecs = delay;
	tr.speed_hz = speed_Hz;
	tr.bits_per_word = bits;


	ret = ioctl(file, SPI_IOC_MESSAGE(1), &tr);
	if (ret < 1)
	{
		printf("ERROR: can't send spi message\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	return 0;
}

/*
 * Read STATUS register
 */
int read_status_reg(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay, uint8_t *status_reg)
{
	int ret;
	uint8_t tx[2];
	uint8_t rx[2];
	struct spi_ioc_transfer tr;


	tx[0] = CMD_READ_STAT;

	/* setup spi transfer struct */
	tr.tx_buf = (unsigned long)&tx;
	tr.rx_buf = (unsigned long)&rx;
	tr.len = 2;
	tr.delay_usecs = delay;
	tr.speed_hz = speed_Hz;
	tr.bits_per_word = bits;


	ret = ioctl(file, SPI_IOC_MESSAGE(1), &tr);
	if (ret < 1)
	{
		printf("ERROR: can't send spi message\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	*status_reg = rx[1];

	return 0;
}

/*
 * Write STATUS register
 */
int write_status_reg(int file, uint8_t bits, uint32_t speed_Hz, uint16_t delay, uint8_t status_reg)
{
	int ret;
	uint8_t tx[2];
	uint8_t rx[2];
	struct spi_ioc_transfer tr;


	tx[0] = CMD_WRITE_STAT;
	tx[1] = status_reg;

	/* setup spi transfer struct */
	tr.tx_buf = (unsigned long)&tx;
	tr.rx_buf = (unsigned long)&rx;
	tr.len = 2;
	tr.delay_usecs = delay;
	tr.speed_hz = speed_Hz;
	tr.bits_per_word = bits;


	ret = ioctl(file, SPI_IOC_MESSAGE(1), &tr);
	if (ret < 1)
	{
		printf("ERROR: can't send spi message\n");
		exit(EXIT_FAILURE); /* Terminate process. All open stdio streams are flushed and closed. */
	}

	return 0;
}
